"""Support for Alembic SQL migrations.

1. Each Python package needs to get its own alembic_history table

2. In "objects to consider" function we resolve the object's package and compare whether it is the package for which we try to create migrations for

3. All migration scripts live inside the package, in alembic/ folder next to setup.py

"""

# Standard Library
import logging
import typing as t

# SQLAlchemy
from alembic import context
from sqlalchemy.ext.declarative.clsregistry import _ModuleMarker
from sqlalchemy.testing.schema import Table

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.devop.cmdline import setup_console_logging
from websauna.system.model.meta import Base


#: Defined later because of initialization order
logger = None


def get_migration_table_name(allowed_packages: t.List, current_package_name) -> str:
    """Convert Python package name to migration table name."""

    package_name = allowed_packages[0]
    if package_name == "all":
        package_name = current_package_name

    table = package_name.replace(".", "_").lower()
    return "alembic_history_{}".format(table)


def get_class_by_table(table: Table) -> type:
    """Return class reference mapped to table.

    :param tablename: String with name of table.
    :return: Class reference or None.
    """
    for c in Base._decl_class_registry.values():
        model_table = getattr(c, "__table__", None)

        # We need to compare by names, because drop declarations use different table objects and eq comparison fails
        if model_table is not None:
            if model_table.name == table.name:
                return c

    return None


def run_migrations_offline(url, target_metadata, version_table, include_object):
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        version_table=version_table,
        include_object=include_object,
        compare_types=True)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online(engine, target_metadata, version_table, include_object):
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    connectable = engine

    with connectable.connect() as connection:

        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=version_table,
            include_object=include_object,
            compare_type=True,  # http://stackoverflow.com/a/17176843/315168
        )

        with context.begin_transaction():
            context.run_migrations()


def consider_class_for_migration(klass: type, allowed_packages: t.List) -> bool:

    if allowed_packages[0] == "all":
        return True

    if any(klass.__module__.startswith(pkg) for pkg in allowed_packages):
        return True

    logger.debug("Skipping SQLAlchemy model %s as out of scope for packages %s", klass, allowed_packages)
    return False


def consider_module_for_migration(mod: str, allowed_packages: t.List) -> bool:
    """
    :param mod: Dotted name to a module
    :param allowed_packages: List of allowed packages
    :return:
    """

    if allowed_packages[0] == "all":
        return True

    return any(mod.startswith(pkg) for pkg in allowed_packages)


def get_sqlalchemy_metadata(allowed_packages: t.List):
    """Get the SQLAlchemy MetaData instance which contains declarative tables only from a certain Python packagek.

    We get all tables which have been registered against the current Base model. Then we grab Base.metadata and drop out all tables which we think are not part of our migration run.
    """

    allowed_tables = []

    # Include all SQLAlchemy models in the local namespace
    for name, klass in Base._decl_class_registry.items():

        if isinstance(klass, _ModuleMarker):
            continue

        if not consider_class_for_migration(klass, allowed_packages):
            continue

        allowed_tables.append(klass.__table__)

    # Remove metadata table registrations which did not below to the package
    metadata = Base.metadata
    for table in list(metadata.tables.values()):
        if table not in allowed_tables:
            metadata.remove(table)

    return metadata


def parse_allowed_packages(current_package):
    """Return list of package names we want to consider from Alembic extra arguments."""

    # Extra -x arguments passed to Alembic
    extra = context.get_x_argument(as_dictionary=True)

    # XXX: Make this a proper command line switch when writing more refined Alembic front end
    packages = extra.get("packages", current_package)
    packages = packages.split(",")

    if packages == "all":
        # Force Alembic to consider all packages
        logger.info("Considering migrations for models in all Python packages")
    else:
        logger.info("Considering migrations for models in Python packages %s", packages)

    return packages


def run_alembic(package: str):
    """Alembic env.py script entry point for Websauna application.

    Initialize the application, load models and pass control to Alembic migration handler.

    :param package: String of the Python package name whose model the migration concerns. E.g. the name of the current Websauna package.
    """
    global logger
    global version_table

    current_package = package

    # this is the Alembic Config object, which provides
    # access to the values within the .ini file in use.
    config = context.config

    # This was -c passed to ws-alembic command
    config_file = config.config_file_name

    # Alembic always uses console logging
    setup_console_logging(logging.INFO)

    # Load the WSGI application, etc.
    request = init_websauna(config_file)
    engine = request.dbsession.get_bind()

    # Delay logger creation until we have initialized the logging system
    logger = logging.getLogger(__name__)

    # Extract database connection URL from the settings
    url = request.registry.settings["sqlalchemy.url"]

    allowed_packages = parse_allowed_packages(current_package)

    #: Pull out MetaData instance for the system
    target_metadata = get_sqlalchemy_metadata(allowed_packages)

    # Each package needs to maintain its own alembic_history table
    version_table = get_migration_table_name(allowed_packages, current_package)

    def include_object(object, name, type_, reflected, compare_to):
        """
        """
        # Try to figure out smartly table from different object types
        if type_ in ("index", "column", "foreign_key_constraint", "unique_constraint"):
            table_name = object.table.name
            table = object.table
        elif type_ == "table":
            table_name = object.name
            table = object
        else:
            raise RuntimeError("Don't know how to check type for migration inclusion list: {}".format(type_))

        model = get_class_by_table(table)
        if not model:
            # Don't know what's this... let's skip
            logger.info("No model available: %s %s %s", object, type_, table_name)
            return False

        module = model.__module__

        # XXX: This is not very beautiful check but works for now
        return consider_module_for_migration(module, allowed_packages)

    if context.is_offline_mode():
        run_migrations_offline(url, target_metadata, version_table, include_object)
    else:
        logger.info("Starting online migration engine on database connection {} version history table {}".format(engine, version_table))
        run_migrations_online(engine, target_metadata, version_table, include_object)

    # TODO: If a migration file is written, post-edit it and add websauna import

    logger.info("All done")



"""Display yes/no confirmation pages before taking action."""
# Standard Library
import typing as t

# Pyramid
from pyramid.httpexceptions import HTTPBadRequest
from pyramid.session import check_csrf_token

# Websauna
from websauna.system.http import Request


class Choice:
    """Present one interstitial choice button given to a user."""

    def __init__(self, label: str, callback: t.Callable, id: str=None, css_class=None, icon_class=None):
        """
        :param label: Human readable label
        :param callback: Function called when this choice is made. Takes Must return
        :param id: Form button id. If not given derive from ``callback``.
        :param css_class: Extra CSS clases for button, like ``btn-danger``
        :param icon_class: CSS classes for ``<i>`` icon tag, like ``fa fa-remove``
        :return:
        """
        self.label = label
        self.callback = callback
        self.id = id or self.callback.__name__
        self.css_class = css_class


def process_interstitial(request: Request, choices: t.List[Choice], *args, **kwargs):
    """Check if user pressed any of the buttons on form and the choice accordingly.

    For example use case see :py:class:`websauna.system.crud.views.Delete`.

    :param args: Passed to choice callback
    :param kwargs: Passed to choice callback
    :return: HTTP response given by a choice callback
    """
    assert request.method == "POST"

    # Force CSRF check always
    check_csrf_token(request)

    for c in choices:
        if c.id in request.POST:
            return c.callback(*args, **kwargs)

    raise HTTPBadRequest("Unknown choice made")



"""Testing utility functions."""

# Standard Library
import subprocess
import time
import typing as t

# Pyramid
import transaction
from pyramid.interfaces import IRequest
from pyramid.registry import Registry
from pyramid.session import signed_deserialize
from zope.interface import implementer

# SQLAlchemy
from sqlalchemy.orm import Session

from pyramid_redis_sessions import RedisSession
from pyramid_redis_sessions import get_default_connection
from selenium.webdriver.remote.webdriver import WebDriver
from splinter.driver import DriverAPI

# Websauna
from websauna.system.user.interfaces import IPasswordHasher
from websauna.system.user.models import User
from websauna.system.user.utils import get_site_creator
from websauna.utils.time import now


EMAIL = "example@example.com"

#: Unit testing default password
PASSWORD = "ToholamppiMadCowz585"


def create_user(dbsession: Session, registry: Registry, email: str=EMAIL, password: str=PASSWORD, admin: bool=False) -> User:
    """A helper function to create normal and admin users for tests.

    Example:

    .. code-block:: python

        import transaction
        from websauna.tests.utils import create_user


        def test_some_stuff(dbsession, registry):

            with transaction.manager:
                u = create_user(registry)
                # Do stuff with new user



    :param email: User's email address. If inot given use unit testing default.

    :param password: Password as plain text. If not given use unit testing default.

    :param admin: If True run :py:class:`websauna.system.user.usermixin.SiteCreator` login and set the user to admin group.
    """

    user = User(email=email)

    if password:
        hasher = registry.getUtility(IPasswordHasher)
        user.hashed_password = hasher.hash_password(password)

    user.user_registration_source = "dummy"
    dbsession.add(user)
    dbsession.flush()
    user.username = user.generate_username()
    user.activated_at = now()

    assert user.can_login()

    # First user, make it admin
    if admin:
        site_creator = get_site_creator(registry)
        site_creator.init_empty_site(dbsession, user)

    return user


def create_logged_in_user(dbsession: Session, registry: Registry, web_server: str, browser: DriverAPI, admin: bool=False, email: str=EMAIL, password: str=PASSWORD):
    """For a web browser test session, creates a new user and log it in inside the test browser."""
    # Catch some common argument misordering issues
    assert isinstance(registry, Registry)
    assert isinstance(web_server, str)

    with transaction.manager:
        create_user(dbsession, registry, admin=admin, email=email, password=password)

    b = browser
    b.visit("{}/{}".format(web_server, "login"))

    assert b.is_element_present_by_css("#login-form")

    b.fill("username", email)
    b.fill("password", password)
    b.find_by_name("login_email").click()

    # After login we log out link to confirm login has succeeded
    assert b.is_element_present_by_css("#nav-logout")


def wait_until(callback: t.Callable, expected: object, deadline=1.0, poll_period=0.05):
    """A helper function to wait until a variable value is set (in another thread).

    This is useful for communicating between Selenium test driver and test runner main thread.

    :param callback: t.Callable which we expect to return to ``expected`` value.
    :param deadline: Seconds how long we are going to wait max
    :param poll_period: Sleep period between check attemps
    :return: The final value of callback
    """
    expires = time.time() + deadline
    while time.time() < expires:
        val = callback()
        if val == expected:
            return val
        time.sleep(poll_period)

    raise AssertionError("Callback {} did not return in expected value {} within {} seconds".format(callback, expected, deadline))


def login(web_server: str, browser: DriverAPI, email=EMAIL, password=PASSWORD):
    """Login user to the website through the test browser."""
    b = browser
    b.visit(web_server)
    assert b.is_element_visible_by_css("#nav-sign-in"), "Could not see login button"
    b.find_by_css("#nav-sign-in").click()
    b.fill("username", email)
    b.fill("password", password)
    b.find_by_name("login_email").click()
    assert b.is_element_visible_by_css("#nav-logout")


def get_session_from_webdriver(driver: WebDriver, registry: Registry) -> RedisSession:
    """Extract session cookie from a Selenium driver and fetch a matching pyramid_redis_sesssion data.

    Example::

        def test_newsletter_referral(dbsession, web_server, browser, init):
            '''Referral is tracker for the newsletter subscription.'''

            b = browser
            b.visit(web_server + "/newsletter")

            with transaction.manager:
                r = ReferralProgram()
                r.name = "Foobar program"
                dbsession.add(r)
                dbsession.flush()
                ref_id, slug = r.id, r.slug

            # Inject referral data to the active session. We do this because it is very hard to spoof external links pointing to localhost test web server.
            session = get_session_from_webdriver(b.driver, init.config.registry)
            session["referral"] = {
                "ref": slug,
                "referrer": "http://example.com"
            }
            session.to_redis()

            b.fill("email", "foobar@example.com")
            b.find_by_name("subscribe").click()

            # Displayed as a message after succesful form subscription
            assert b.is_text_present("Thank you!")

            # Check we get an entry
            with transaction.manager:
                assert dbsession.query(NewsletterSubscriber).count() == 1
                subscription = dbsession.query(NewsletterSubscriber).first()
                assert subscription.email == "foobar@example.com"
                assert subscription.ip == "127.0.0.1"
                assert subscription.referral_program_id == ref_id
                assert subscription.referrer == "http://example.com"

    :param driver: The active WebDriver (usually ``browser.driver``)

    :param registry: The Pyramid registry (usually ``init.config.registry``)
    """

    # Decode the session our test browser is associated with by reading the raw session cookie value and fetching the session object from Redis
    secret = registry.settings["redis.sessions.secret"]

    session_cookie = driver.get_cookie("session")["value"]
    session_id = signed_deserialize(session_cookie, secret)

    class MockRequest:
        def __init__(self, registry):
            self.registry = registry

    # Use pyramid_redis_session to get a connection to the Redis database
    redis = get_default_connection(MockRequest(registry))

    session = RedisSession(redis, session_id, new=False, new_session=None)

    return session


def logout(web_server: str, browser: DriverAPI):
    """Log out the current user from the test browser."""
    browser.find_by_css("#nav-logout").click()
    assert browser.is_element_present_by_css("#msg-logged-out")


def make_dummy_request(dbsession: Session, registry: Registry) -> IRequest:
    """Creates a non-functional HTTP request with registry and dbsession configured.

    Useful for crafting requests with custom settings

    See also :func:`make_routable_request`.
    """

    @implementer(IRequest)
    class DummyRequest:
        pass

    _request = DummyRequest()
    _request.dbsession = dbsession
    _request.user = None
    _request.registry = registry

    return _request


def print_subprocess_fail(worker, cmdline):
    print("{cmdline} output:".format(cmdline=cmdline))
    print(worker.stdout.read().decode("utf-8"))
    print(worker.stderr.read().decode("utf-8"))


def execute_command(cmdline: t.List, folder: str, timeout=5.0):
    """Run a command in a specific folder."""
    worker = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=folder)

    try:
        worker.wait(timeout)
    except subprocess.TimeoutExpired as e:
        print_subprocess_fail(worker, cmdline)
        raise AssertionError("execute_command did not properly exit") from e

    if worker.returncode != 0:
        print_subprocess_fail(worker, cmdline)
        raise AssertionError("scaffold command did not properly exit: {}".format(" ".join(cmdline)))

    return worker.returncode



"""
These tests aim to ensure that websauna's ``simple_route`` can be used in the
same way as pyramid's ``view_config``. Tests are inspired by pyramid's
`viewconfig documentation`_ and aim to check that the provided examples of the
use of ``view_config`` are also applicable to ``simple_route``.

.. _`viewconfig documentation`: https://docs.pylonsproject.org/projects/pyramid/en/latest/narr/viewconfig.html#adding-view-configuration-using-the-view-config-decorator

"""
# Pyramid
from pyramid import testing

import pytest
from webtest import TestApp


@pytest.fixture(scope='module')
def config():
    with testing.setUp() as config:
        config.scan('websauna.system.tests.viewconfig.simple_route_tests_views')
        yield config


@pytest.fixture(scope='module')
def testapp(config):
    yield TestApp(config.make_wsgi_app())


def test_registered_routes(config):
    expected_routes = [
        'ClassAsAView',
        'MyViewClass.amethod',
        'MyViewClassAttr.amethod',
        'edit',
        'change',
        'foobar',
        'view_callable',
    ]
    actual_routes = [i.name for i in config.get_routes_mapper().get_routes()]
    assert expected_routes == actual_routes


def test_view_callable(testapp):
    res = testapp.get('/path/to/view')
    assert res.body == b'Hello'


def test_route_items(testapp):
    res = testapp.post('/path/to/view/Joe')
    assert res.body == b'{"message": "Hello Joe"}'


def test_class_as_a_view(testapp):
    res = testapp.get('/class-as-a-view')
    assert res.body == b'class-as-a-view'


def test_muliple_simple_route_decorators(testapp):
    res = testapp.get('/edit')
    assert res.body == b'edited!'
    res = testapp.get('/change')
    assert res.body == b'edited!'


def test_method_of_a_class(testapp):
    res = testapp.get('/amethod')
    assert res.body == b'amethod'


def test_class_method_by_attr(testapp):
    res = testapp.get('/class-attr')
    assert res.body == b'class-attr'


def test_registered_routes_docs(config):
    config.get_routes_mapper()
    request = testing.DummyRequest()
    assert request.route_url('view_callable') == 'http://example.com/path/to/view'
    assert request.route_url('foobar', arg='123') == 'http://example.com/path/to/view/123'
    assert request.route_url('edit') == 'http://example.com/edit'
    assert request.route_url('change') == 'http://example.com/change'
    assert request.route_url('ClassAsAView') == 'http://example.com/class-as-a-view'
    assert request.route_url('MyViewClass.amethod') == 'http://example.com/amethod'
    assert request.route_url('MyViewClassAttr.amethod') == 'http://example.com/class-attr'



"""Celery async task tests."""
# Standard Library
import os
import subprocess
import time

# Pyramid
import transaction

import pytest
from flaky import flaky

# Websauna
from websauna.system.devop.cmdline import init_websauna
from websauna.system.task.celery import get_celery
from websauna.system.user.models import User

from websauna.tests import demotasks


@pytest.fixture(scope='module')
def task_ini_file():
    return os.path.join(os.path.dirname(__file__), "task-test.ini")


@pytest.fixture(scope='module')
def task_app_request(task_ini_file, ini_settings):
    request = init_websauna(task_ini_file)

    if ini_settings["sqlalchemy.url"].startswith("sqlite://"):
        pytest.skip("These tests are run only for PostgreSQL database")

    return request


@pytest.fixture()
def demo_user(request, dbsession):
    """Create a database object asyncronously manipulated."""

    with transaction.manager:
        # Do a dummy database write
        u = User(username="test", email="test@example.com")
        dbsession.add(u)


@pytest.fixture(scope='module')
def celery_worker(request, task_ini_file):
    """py.test fixture to shoot up Celery worker process to process our test tasks when scheduled."""

    # Uncomment this and run ws-celery from command line for debug
    # ws-celery ws://websauna/system/tests/task-test.ini -- worker --loglevel=debug
    # return
    cmdline = "ws-celery ws://{ini_file} -- worker".format(ini_file=task_ini_file)

    # logger.info("Running celery worker: %s", cmdline)

    worker = subprocess.Popen(cmdline, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    time.sleep(4.0)

    worker.poll()
    if worker.returncode is not None:
        print(worker.stdout.read().decode("utf-8"))
        print(worker.stderr.read().decode("utf-8"))
        raise AssertionError("Celery worker process did not start up: {}".format(cmdline))

    def teardown():
        worker.terminate()
        # XXX: Hard to capture this only on failure for now
        # print(worker.stdout.read().decode("utf-8"))
        # print(worker.stderr.read().decode("utf-8"))

    request.addfinalizer(teardown)

    return worker.pid


@flaky
def test_transaction_aware_task_success(celery_worker, task_app_request, dbsession, demo_user):
    """Transaction aware tasks works in eager mode.."""
    with transaction.manager:
        # Do a dummy database write
        u = dbsession.query(User).first()

        demotasks.modify_username.apply_async([u.id], tm=transaction.manager)
        # Let the task travel in Celery queue
        time.sleep(2.0)

        # Task should not fire unless we exit the transaction
        assert u.username != "set by celery"

    # Let the transaction commit
    time.sleep(0.5)

    # Task has now fired after transaction was committed
    with transaction.manager:
        u = dbsession.query(User).get(1)
        assert u.username == "set by celery"


@flaky
def test_transaction_manager_abort(celery_worker, task_app_request, dbsession, demo_user):
    """Test that transaction are not executed to schedule if transaction manager aborts."""

    try:
        # Tasks do not execute if transaction is never commit because of exception
        with transaction.manager:
            assert dbsession.query(User).count() == 1
            assert dbsession.query(User).get(1).username == "test"
            demotasks.modify_username.apply_async(args=[1], tm=transaction.manager)
            raise RuntimeError("aargh")
    except RuntimeError:
        pass

    # Let the transaction commit (which should not happen)
    time.sleep(0.5)

    # Because of exception, Celery never fires
    with transaction.manager:
        u = dbsession.query(User).get(1)
        assert u.username == "test"


@flaky
def test_task_throws_exception(celery_worker, task_app_request, dbsession, demo_user):
    """Test that transaction aware task does not commit if exception is thrown.."""

    with transaction.manager:
        u = dbsession.query(User).first()
        demotasks.modify_username_abort_by_exception.apply_async(args=[u.id], tm=transaction.manager)

    # Let the Celery task execute
    time.sleep(0.5)

    with transaction.manager:
        # Value should not have changed
        u = dbsession.query(User).first()
        assert u.username == "test"


@flaky
def test_manual_transaction(celery_worker, task_app_request, dbsession, demo_user):
    """Manual transaction lifecycles within task work."""

    with transaction.manager:
        # Do a dummy database write
        u = dbsession.query(User).first()

        demotasks.modify_username_manual_transaction.apply_async(kwargs={"user_id": u.id}, tm=transaction.manager)
        # Let the task travel in Celery queue
        time.sleep(1.0)

        # Task should not fire unless we exit the transaction
        assert u.username != "set by celery"

    # Let the transaction commit
    time.sleep(0.5)

    # Task has now fired after transaction was committed
    with transaction.manager:
        u = dbsession.query(User).first()
        assert u.username == "set by celery"


@flaky
def test_eager(celery_worker, task_app_request, dbsession, demo_user):
    """When in eager mode, transactions are executed properly.."""

    celery = get_celery(task_app_request.registry)
    celery._conf["task_always_eager"] = True

    try:

        # Try RetryableTransactionTask in eager mode
        with transaction.manager:
            # Do a dummy database write
            u = dbsession.query(User).first()

            demotasks.modify_username.apply_async(args=[u.id], tm=transaction.manager)
            # Task should not execute until TM commits
            assert u.username != "set by celery"

        # TM commits the new result should be instantly available

        # Task has now fired after transaction was committed
        with transaction.manager:
            u = dbsession.query(User).first()
            assert u.username == "set by celery"

        # Let's test ScheduleOnCommitTask with manually managed transaction
        with transaction.manager:
            u = dbsession.query(User).first()
            u.username = "foobar"
            demotasks.modify_username_manual_transaction.apply_async(args=[u.id], tm=transaction.manager)

        # ScheduledOnCommitTask should have finished now
        with transaction.manager:
            u = dbsession.query(User).first()
            assert u.username == "set by celery"

    finally:
        celery._conf["task_always_eager"] = False



# Standard Library
import typing as t

# Pyramid
from pyramid.request import Request
from zope.interface import Interface

# Websauna
from websauna.system.core.interfaces import IRoot
from websauna.system.core.traversal import Resource


def get_human_readable_resource_name(resource: Resource) -> str:
    """Extract human-readable name of the resource for breadcrumbs."""
    # TODO: Add adapter implementation here
    if hasattr(resource, "get_title"):
        return resource.get_title()
    if hasattr(resource, "title"):
        return resource.title
    return resource.__name__


def get_breadcrumbs(context: Resource, request: Request, root_iface: type=None, current_view_name=None, current_view_url=None) -> t.List:
    """Create breadcrumbs path data how to get to this resource from the root.

    Traverse context :class:`Resource` up to the root resource in the reverse order. Fill in data for rendering
    Bootstrap breacrumbs title. Each traversed resource must provide ``get_title()`` method giving a human readable title for the resource.

    :param current_view_name: Optional user visible name of the current view for the bottom most resource.

    :param current_view_url: Full URL to the current view

    :param root_iface: If you want to traverse only subset of elements and stop a certain parent, optional root can be marked with an interface. If not given assume :class:`websauna.system.core.interfaces.IRoot`. Optionally traversing is terminated when reaching ``None`` as the ``__parent__`` pointer.

    :return: List of {url, name, resource} dictionaries
    """
    elems = []
    if not root_iface:
        root_iface = IRoot
    assert issubclass(root_iface, Interface), "Traversing root must be declared by an interface, got {}".format(root_iface)

    # Looks like it is not possible to dig out the matched view from Pyramid request,
    # so we need to explicitly pass it if we want it to appear in URL
    if current_view_name:
        assert current_view_url
        elems.append(dict(url=current_view_url, name=current_view_name))

    while context and not root_iface.providedBy(context):
        if not hasattr(context, "get_title"):
            raise RuntimeError("Breadcrumbs part missing get_title(): {}".format(context))

        elems.append(dict(url=request.resource_url(context), name=get_human_readable_resource_name(context), resource=context))

        if not hasattr(context, "__parent__"):
            raise RuntimeError("Broken traverse lineage on {}, __parent__ missing".format(context))

        if not isinstance(context, Resource):
            raise RuntimeError("Lineage has item not compatible with breadcrums: {}".format(context))

        context = context.__parent__

    # Add the last (root) element
    entry = dict(url=request.resource_url(context), name=get_human_readable_resource_name(context), resource=context)
    elems.append(entry)
    elems.reverse()

    return elems



"""Sitemap generation helpers."""
# Standard Library
import abc
import typing as t

# Pyramid
from pyramid.config import Configurator
from pyramid.interfaces import IAuthorizationPolicy
from pyramid.interfaces import IRouteRequest
from pyramid.interfaces import ITraverser
from pyramid.registry import Introspectable
from pyramid.router import Router
from pyramid.scripts.proutes import _get_pattern
from pyramid.security import Everyone
from pyramid.static import static_view
from pyramid.traversal import ResourceTreeTraverser
from pyramid.urldispatch import Route

# Websauna
from websauna.system.core.interfaces import IContainer
from websauna.system.core.traversal import Resource
from websauna.system.http import Request
from websauna.system.http.utils import make_routable_request


class SitemapItem(abc.ABC):
    """Present an item apperearing in the sitemap.

    Pass information about the sitemap entry for the template.
    """

    def __init__(self, changefreq=None, priority=None, lastmod=None):
        self._changefreq = changefreq
        self._priority = priority
        self._lastmod = lastmod

    @abc.abstractmethod
    def location(self, request):
        """Resolve the full URL of the this item in the sitemap.

        :return: URL as a string
        """
        raise NotImplementedError()

    def changefreq(self, request):
        """Return sitemap changefreq string or None if tag not present."""
        return self._changefreq

    def priority(self, request):
        """Return sitemap priority string or None if priority tag string not present."""
        return self._priority

    def lastmod(self, request):
        """Return sitemap lastmod string or None if tag not present."""
        return self._lastmod


class RouteItem(SitemapItem):
    """Add a static Pyramid URL dispatched route to the sitemap. """

    def __init__(self, route_name, changefreq=None, priority=None, lastmod=None, **kwargs):
        super(RouteItem, self).__init__(changefreq, priority, lastmod)
        self.route_name = route_name
        self.kwargs = kwargs

    def location(self, request):
        return request.route_url(self.route_name, **self.kwargs)


class TraverseItem(SitemapItem):
    """Add a travered resource view to the sitemap."""

    def __init__(self, context: Resource, view_name, changefreq=None, priority=None, lastmod=None, **kwargs):
        super(TraverseItem, self).__init__(changefreq, priority, lastmod)
        self.context = context
        self.view_name = view_name

    def location(self, request):
        return request.resource_url(self.context, self.view_name)


class Sitemap:
    """Sitemap helper.
    See :ref:`Sitemap <sitemap>`.

    .. _sitemap:
    """

    def __init__(self):
        self.items = []
        self.generators = []

    def add_item(self, item):
        assert isinstance(item, SitemapItem)
        self.items.append(item)

    def add_generator(self, generator):
        """
        :param generator: Yields SitemapItem instances
        """
        self.generators.append(generator)

    def urls(self):
        """Return an iterable which goes through all SitemapItem objects in this Sitemap."""
        for item in self.items:
            yield item

        for generator in self.generators:
            yield from generator()

    def render(self, context, request):
        """Render the sitemap.

        :return: dict of information for the templates {urlset: SitemapItem iterator}
        """
        request.response.content_type = "application/xml"
        return dict(urlset=self.urls())


class ReflectiveSitemapBuilder:
    """Scan all registered routes and traversable resources and build sitemap from them automatically.

    This will read route configuration and build sitemap for

    * All routes without parameter

    * All traversable endpoints that implement :py:class:`websauna.system.core.interfaces.IContainer` protocol

    * GET accessible views

    This method might not yet work for more advanced view configuration use cases. Check :py:mod:`websauna.tests.sitemapsamples` for covered use cases.

    See :ref:`Sitemap <sitemap>` for examples.
    """

    def __init__(self, request: Request):
        self.request = request
        self.sitemap = Sitemap()

    def get_mapper(self):
        config = Configurator(registry=self.request.registry)
        return config.get_routes_mapper()

    def is_parameter_free_route(self, route_spec: str):
        # no {param}, no traverse/*
        return ("{" not in route_spec) and ("*" not in route_spec)

    def is_traversable_sitemap_route(self, route: Route):
        """Is this route such that it 1) is traversable 2) can be added to sitemap"""

        if not route.factory:
            # /notebook/*remainder
            # static view
            return False

        return "*" in route.pattern

    def is_get_requestable(self, view_data: dict):

        if not view_data.get("request_methods"):
            return True

        return "GET" in view_data["request_methods"]

    def is_anonymous(self, view_data: dict):

        # TODO: This does not handle the case where
        # multiple derivates are stacked
        derived = view_data.get("derived_callable")
        if derived:
            if getattr(derived, "__permission__", None):
                return False

        return True

    def is_static(self, view_data: dict):
        return isinstance(view_data.get("callable"), static_view)

    def is_public_get_view(self, view_data: dict):
        """Check if view can be publicly accessed

        :param view_data: Introspected view data as dict
        """

        if not self.is_get_requestable(view_data):
            return False

        if not self.is_anonymous(view_data):
            return False

        return True

    def is_included(self, view_data: dict, context: t.Optional[Resource], request: Request):
        """Check if sitemap conditions allow to include this item."""

        callable = view_data.get("callable")
        if callable is None:
            # This is route like admin_home without callable, only traversing nesting
            return True

        # TODO: Not sure if we need to peek through callable decorator stack
        sitemap_data = getattr(callable, "_sitemap_data", None)
        if sitemap_data is None:
            return True

        # Hardcoded condition
        included = sitemap_data.get("include", None)
        if included is not None:
            return included

        condition = sitemap_data["condition"]
        return condition(context, request)

    def is_good_route_item(self, name, pattern, view_data):
        """Check conditions if routed view can be added in the sitemap."""

        # view_data.keys()
        # dict_keys(['mapper', 'header', 'check_csrf', 'predicates', 'containment', 'attr', 'route_name', 'request_methods', 'order', 'xhr', 'path_info', 'phash', 'callable', 'decorator', 'name', 'request_param', 'context', 'require_csrf', 'derived_callable', 'accept', 'http_cache', 'match_param'])

        if not self.is_parameter_free_route(pattern):
            return False

        if not self.is_public_get_view(view_data):
            return False

        if not self.is_included(view_data, None, self.request):
            return False

        return True

    def has_public_view_acl(self, context: Resource):
        """Check if ACL for the resource is publicly viewable.

        View permission must be either missing or pyramid.security.Everyone.
        """
        policy = self.request.registry.queryUtility(IAuthorizationPolicy)

        # view permission is set on Root object or overridden in resource hierarchy __ACL__
        principals = policy.principals_allowed_by_permission(context, "view")
        return Everyone in principals

    def add_route_item(self, name, pattern, view_data):
        """Add one route item to the table.

        Override for custom sitemap parameters (changefreq, etc.).
        """
        self.sitemap.add_item(RouteItem(name))

    def add_traverse_item(self, context: Resource, view_name: str):
        """Add one traverse item to the table.

        Override for custom sitemap parameters (changefreq, etc.).
        """
        self.sitemap.add_item(TraverseItem(context, view_name))

    def build_routes(self):
        """Build all routes without parameters and permissions."""
        mapper = self.get_mapper()
        routes = mapper.get_routes(include_static=False)

        for route in routes:
            data = _get_route_data(route, self.request.registry)
            for name, pattern, view_data in data:

                if self.is_good_route_item(name, pattern, view_data):
                    self.add_route_item(name, pattern, view_data)

    def enumerate_available_views(self, route: Route, context: Resource) -> t.Iterable[Introspectable]:
        """Get list of available views for a given resource."""
        introspector = self.request.registry.introspector
        views = introspector.get_category("views")

        # We can have multiple views with the same name for the same resource because of inheritance.
        available_views = {}

        for v in views:
            spectable = v["introspectable"]  # pyramid.registry.Introspectable

            # See discrim_func() in add_view()
            cat, intr_context_cls, name, route_name, phash = spectable.discriminator

            if not intr_context_cls:
                continue

            if route_name != route.name:
                # On a different route
                continue

            if isinstance(context, intr_context_cls):
                # TODO: Assumes that the later defined views have more accurate context scope.
                # Not sure if this holds true all the time.
                available_views[name] = v["introspectable"]

        return available_views.values()

    def get_traverse_endpoint_context(self, router: Router, route: Route) -> Resource:
        """Get root object for a traversable route.

        E.g. resolve /container/* to a SampleContainer context.
        """

        # chop off last part /container/*traverse'
        start_path = "/".join(route.pattern.split("/")[0:-2]) + "/"
        sample_request = make_routable_request(path=start_path, registry=self.request.registry)

        root = route.factory(sample_request)

        traverser = self.request.registry.queryAdapter(root, ITraverser)
        # We are about to traverse and find a context
        if traverser is None:
            traverser = ResourceTreeTraverser(root)

        # {'virtual_root': <websauna.tests.sitemapsamples.SampleContainer object at 0x104656f98>, 'traversed': (), 'root': <websauna.tests.sitemapsamples.SampleContainer object at 0x104656f98>, 'virtual_root_path': (), 'view_name': 'container', 'subpath': (), 'context': <websauna.tests.sitemapsamples.SampleContainer object at 0x104656f98>}
        tdict = traverser(sample_request)
        context = tdict["context"]
        return context

    def recurse_traversable(self, router: Router, route: Route, context: Resource):
        """Walk through traversable hierarchy.

        For each context iterate available views and add to sitemap.
        """

        if not self.has_public_view_acl(context):
            # This resource limits view permission to subgroup and is not public. E.g. /admin/*
            return

        # Add all views for this leaf
        for view_data in self.enumerate_available_views(route, context):
            if self.is_public_get_view(view_data) and self.is_included(view_data, context, self.request):
                self.add_traverse_item(context, view_data["name"])

        # Recurse to children
        if IContainer.providedBy(context):
            for name, child in context.items():
                self.recurse_traversable(router, route, child)

    def build_traverse_trees(self):
        """Build all traversed hierarchies."""
        mapper = self.get_mapper()
        router = Router(self.request.registry)

        routes = mapper.get_routes(include_static=False)

        for route in routes:
            if not self.is_traversable_sitemap_route(route):
                continue

            root_context = self.get_traverse_endpoint_context(router, route)
            self.recurse_traversable(router, route, root_context)

    def build(self):
        """Iterate through all public routes and traversable items and add them to the sitemap."""
        self.build_routes()
        self.build_traverse_trees()

    def get_sitemap(self) -> Sitemap:
        """Get ready sitemap after build."""
        return self.sitemap

    @classmethod
    def render(cls, context, request):
        """Render the sitemap.

        TODO: I don't like that Sitemap.render is little bit different from this one.

        :return: dict of information for the templates {urlset: SitemapItem iterator}
        """
        reflective_builder = cls(request)
        reflective_builder.build()
        map = reflective_builder.get_sitemap()
        return map.render(context, request)


def _get_route_data(route, registry):
    """Iterate all non static views for a route.

    :param route: Route introspection data
    :param registry:
    :yield: (route name, route pattern, view introspection data) tuples
    """

    pattern = _get_pattern(route)

    request_iface = registry.queryUtility(
        IRouteRequest,
        name=route.name
    )

    route_intr = registry.introspector.get(
        'routes', route.name
    )

    if request_iface is None:
        return

    if route_intr.get('static', False) is True:
        return

    view_intr = registry.introspector.related(route_intr)

    if view_intr:
        for view in view_intr:
            yield route.name, pattern, view


def include_in_sitemap(include: t.Optional[bool]=None, condition: t.Optional[t.Callable]=None):
    """A function decorator to determine if a view should be included in the automatically generated sitemap.

    You need to give either ``include`` argument or ``condition``. If this view decorator is not present the view is always included.

    Example of hardcoded condition for an URL dispatch view:

    .. code-block:: python

        from websauna.system.core.sitemap import include_in_sitemap
        from websauna.system.core.route import simple_route


        @simple_route("/skipped_route", route_name="skipped_route")
        @include_in_sitemap(False)
        def skipped_route(request: Request):
            return Response()

    Example of dynamic condition for a traverse view:

    .. code-block:: python

        from websauna.system.core.sitemap import include_in_sitemap
        from pyramid.view import view_config


        def skipped_condition(context, request):
            return False


        @view_config(context=SampleResource, name="skipped_conditional", route_name="sitemap_test")
        @include_in_sitemap(condition=skipped_condition)
        def skipped_conditional(sample_resource: SampleResource, request: Request):
            return Response()

    More complex example where you dynamically look up from the database if the content exist and then have the item in the sitemap:

    .. code-block:: python

        # We reuse this helper function in sitemap condition and actual view
        def get_insight_content(asset_resource):
            asset = asset_resource.asset
            content = asset.asset_content.filter_by(content_type=AssetContentType.insight).one_or_none()
            return content


        @view_config(context=AssetDescription, route_name="network", name="insight", renderer="network/asset_insight.html")
        @include_in_sitemap(lambda c, r: get_insight_content(c) is not None)
        def insight(asset_resource: AssetDescription, request: Request):
            breadcrumbs = get_breadcrumbs(asset_resource, request, current_view_name="Asset insight", current_view_url=request.resource_url(asset_resource, "insight"))
            content = get_insight_content(asset_resource)

            if content:
                body = markdown.markdown(content.text)
            else:
                body = "This research has not yet been published"

            return locals()

    :param include: Either ``True`` or ``False``

    :param condition: callback function (context, request) to called to every item. Traversable views get both route and request, URL dispatch views get only request. Return true if the item should be included in the sitemap.
    """

    assert (include is not None) or (condition is not None), "You need to give either include or condition argument"

    def _outer(view_func):
        # Store data directly on a view function
        view_func._sitemap_data = dict(include=include, condition=condition)
        return view_func

    return _outer



"""Deform lib resource management."""

# Pyramid
import deform.widget
from zope.interface import implementer

# Websauna
from websauna.system.form.interfaces import IFormResources


default_resources = deform.widget.default_resources

# Make sure pick-a-date widget works out of the box
default_resources["pickadate"][None]["css"] = list(default_resources["pickadate"][None]["css"]) + ["websauna.system:static/pickadate-fixes.css"]


@implementer(IFormResources)
class DefaultFormResources:
    """Default form resources implementation.

    If you need to override or manipulate default form resources you can do it by overriding :py:class:`websauna.system.form.interfaces.IFormResources` utility. See :py:meth:`websauna.system.Initializer.configure_forms` for more inforation.
    """

    def get_default_resources(self):
        return default_resources
